/*
 * linux/arch/sh/kernel/setup_shmse.c
 *
 * Copyright (C) 2003 Takashi Kusuda <kusuda-takashi@hitachi-ul.co.jp>
 *
 * SH-Mobile SolutionEngine Support.
 */

#include <linux/config.h>
#include <linux/init.h>
#include <linux/irq.h>

#include <asm/io.h>
#include <asm/hitachi_shmse.h>

void __init init_shmse_IRQ(void)
{
}

void __init setup_shmse(void)
{
}
